/* ======================================================================
     "bx1HRsol.cc"
       - Solve one-dimensional time-independent Schroedinger equation
          by basis function expansion (FS: for Fourier series potential
          and free rotor basis)
             Copyright (c) 2004-2005  by A. Miyoshi, Univ. Tokyo
                         created: Sept. 13, 2004   solHRxp.c Rev. 0.10
                          edited: Nov.  29, 2004   solHRxp.c Rev. 0.11
                     last edited: July  11, 2005   solHRxp.c Rev. 0.12
      created based on solHRxp.c: Oct.  12, 2005
                          edited: Oct.  14, 2005   Rev. 0.10
                     last edited: Jan.   6, 2007   Rev. 0.11
 ------------------------------------------------------------------------
 usage
  $ bx1HRsol basefilename
  - input is read from 'basefilename'.inp

 ====================================================================== */

#include "bx1HRsol.h"

/* ===== main function ================================================ */

int main(int argc, char *argv[]) {
  string basfln;
  int outcnt = OUTnone, fullMxAbQnbr, smallMxAbQnbr, limAbQnbr;
  tempList templ;
  basisFreeRotor *basp;   // - large object cannot be an autovariable

  if (argc != 2) { errms(ERRusage); return 1; }
  basfln = argv[1];
  basp = new basisFreeRotor;
  double egvS[basp->MXNBAS];
  if (read_input(basfln, *basp, templ, outcnt, fullMxAbQnbr) == ERRE)
   { return 1; }
  smallMxAbQnbr = fullMxAbQnbr / 2;
  if (basp->solve(smallMxAbQnbr) == ERRE) { errms(NUMerr); return 1; }
  store_egvals(*basp, egvS);
  if (basp->solve(fullMxAbQnbr) == ERRE) { errms(NUMerr); return 1; }
  if ((limAbQnbr = max_reliable_aqn(*basp, smallMxAbQnbr, egvS)) < 0)
   { return 1; }
  if (write_eigen_values(basfln, *basp, limAbQnbr, fullMxAbQnbr)
   == ERRE) { return 1; }
  if (write_part_funcs(basfln, *basp, limAbQnbr, templ) == ERRE)
   { return 1; }
  if (outcnt & OUTvec) {
    if (write_eigen_vectors(basfln, *basp, limAbQnbr) == ERRE)
     { return 1; }
  }
  if (outcnt & OUTfunc) {
    if (write_eigen_funcs(basfln, *basp, limAbQnbr) == ERRE) { return 1; }
  }
  if (outcnt & OUTfunc) {
    if (write_eigen_funcs(basfln, *basp, limAbQnbr) == ERRE) { return 1; }
  }
  delete basp;
  return 0;
}

/* ----- read input --------------------------------------------------- */

int read_input(string &basfln, basisFreeRotor &basis, tempList &tpls,
 int &outcnt, int &fulmaqn) {
  ifstream inf;
  string inpfln = basfln + ".inp", key;
  vector<string> vs, vsv;
  int maqn_read = false, rotc_read = false, potp_read = false;
  int npp = 0, blkend, noc, ioc, order, symn = 0, iss;
  double rotc, coef;

  inf.open(inpfln.c_str());
  if (!inf) { return errms(FAILopen, inpfln); }
  tpls.clear();
  while (inpTokenKeyValues(inf, key, vs) == NORM) {
    if (key == "maxAbsQN") {
      maqn_read = true; fulmaqn = atoi(vs[0].c_str());
    } else if (key == "rotConst") {
      rotc_read = true; rotc = atof(vs[0].c_str());
    } else if (key == "nuSpinSt") {
      symn = atoi(vs[0].c_str()); if (symn < 1) { return errms(INVnuss); }
      if (vs.size() <= symn) { return errms(INVnuss); }
      basis.nSpinW.clear(); 
      for (iss = 0; iss < symn; iss++)
       { basis.nSpinW.push_back(atoi(vs[iss + 1].c_str())); }
    } else if (key == "symNumbr") {
      symn = atoi(vs[0].c_str()); if (symn < 1) { return errms(INVsymn); }
      basis.nSpinW.clear(); 
      for (iss = 0; iss < symn; iss++) { basis.nSpinW.push_back(1); }
    } else if (key == "output") {
      noc = vs.size();
      for (ioc = 0; ioc < noc; ioc++) {
        if (vs[ioc] == "vector") { outcnt |= OUTvec; }
        else if (vs[ioc] == "function") { outcnt |= OUTfunc; }
        else if (vs[ioc] == "all") { outcnt |= OUTall; }
      }
    } else if (key == "tempRange") {
      if (tpls.addRangeVS(vs) == ERRE) { return ERRE; }
    } else if (key == "tempRecipRange") {
      if (tpls.addRecipRangeVS(vs) == ERRE) { return ERRE; }
    } else if (key == "tempList") {
      if (tpls.addListVS(vs) == ERRE) { return ERRE; }
    } else if (key == "potPars{") {
      blkend = false;
      while (!blkend) {
        if (inpToken(inf, vsv) == ERRE) { return errms(READerr); }
        if (vsv[0][0] == '}') { blkend = true; }
        else if (vsv.size() < 2) { return errms(INVpotp); }
        else {
          order = atoi(vsv[0].c_str());
          coef = atof(vsv[1].c_str());
          basis.potE[order] = coef; npp++;
        }
      }
      if (npp > 0) { potp_read = true; basis.potE.finishSetting(); }
    } else if (key != "#") { errms(UNKNkey, key); }
  }
  inf.close();
  if (!maqn_read) { return errms(NOmaqn); }
  if (fulmaqn <= 0) { return errms(INVmaqn); }
  if (!rotc_read) { return errms(NOrotc); }
  if (rotc <= 0.) { return errms(INVrotc); }
  if (!potp_read) { return errms(NOVpotp); }
  if (symn == 0) { basis.nSpinW.clear(); basis.nSpinW.push_back(1); }
  if (tpls.empty()) { tpls.setDefault(); }
  basis.rotB = rotc;
  return NORM;
}

/* ----- store eigenvalues -------------------------------------------- */

void store_egvals(basisFreeRotor &basis, double *egval) {
  int iq, nqn = basis.maxAbsQntNbr * 2 + 1;
  for (iq = 0; iq < nqn; iq++) { egval[iq] = basis.w[iq]; }
}

/* ----- determine the maximum reliable absolute quantum number ------- */

int max_reliable_aqn(basisFreeRotor &basis, int maqnS, double *egvalS)
{
  int iq, limIQS = -1, limAQS, limAQF, maqnF = basis.maxAbsQntNbr;
  int nQnS = maqnS * 2 + 1;
  double fracOK;
  for (iq = 0; iq < nQnS; iq++) {
    if (fabs(basis.w[iq]) >= VAL0THR) {
      if (fabs((egvalS[iq] - basis.w[iq]) / basis.w[iq]) > 0.01)
       { limIQS = iq; break; }
    }
  }
  if (limIQS == -1) { limIQS = nQnS; }
  limAQS = (limIQS - 1) / 2;
  if (limAQS < 0) { limAQS = 0; }
  fracOK = (double)limAQS / maqnS;
  limAQF = (int)(maqnF * fracOK);
  cout << "estimated eigenvalue error is <= 1% at |q.n.| = " << limAQF
   << ".\n";
  if ((limAQF < 3) || (fracOK < 0.01)) { errms(FATALconv); return -1; }
  else if ((limAQF < 10) || (fracOK < 0.03)) { errms(BADconv); }
  return limAQF;
}

/* ----- write eigen values ------------------------------------------- */

int write_eigen_values(string &basfln, basisFreeRotor &basis, int limAQn,
 int maxOutAQn) {
  int iq, Jq, nqn = maxOutAQn * 2 + 1,
   symn = basis.nSpinW.size(), iss, sumSw = 0;
  ofstream ouf;
  string outfln = basfln + "_eigen_values.csv";
  outFormat *ofmt;

  for (iss = 0; iss < symn; iss++) { sumSw += basis.nSpinW[iss]; }
  ouf.open(outfln.c_str());
  if (!ouf) { return errms(FAILopen, outfln); }
  ofmt = new outFormat(&ouf);
  ouf << "number,eigenValue,J,w\n";
  ouf.precision(8);
  for (iq = 0; iq < nqn; iq++) {
    Jq = (iq + 1) / 2; iss = Jq % symn;
    ouf << iq << "," << basis.w[iq] << "," << Jq << ",("
     << basis.nSpinW[iss] << "/" << sumSw << ")";
    if (iq == (limAQn * 2)) { ouf << ",(maximum reliable quantumNbr)"; }
    ouf << endl;
  }
  delete ofmt;
  ouf.close();
  return NORM;
}

/* ----- write eigen vectors ------------------------------------------ */

int write_eigen_vectors(string &basfln, basisFreeRotor &basis,
 int limAQn) {
  int iq, Jq, nqn = limAQn * 2 + 1, qb, ib;
  int nPages = (nqn - 1) / 125 + 1, jp, qbegin, qlast;
  ofstream ouf;
  string outfln;
  outFormat *ofmt;

  for (jp = 0; jp < nPages; jp++) {
    qbegin = jp * 125; qlast = jp * 125 + 124;
    if (qlast >= nqn) { qlast = nqn - 1; }
    outfln = basfln + "_eigen_vectors" + format(jp, 2, '0') + ".csv";
    ouf.open(outfln.c_str());
    if (!ouf) { return errms(FAILopen, outfln); }
    ofmt = new outFormat(&ouf);
    ouf << "eigen_values"; ouf.precision(8);
    for (iq = qbegin; iq <= qlast; iq++)
     { ouf << "," << basis.w[iq] << ","; }
    ouf << "\neigen_vectors";
    for (iq = qbegin; iq <= qlast; iq++) { ouf << ",(real),(imag)"; }
    ouf << "\nJ(basis)";
    for (iq = qbegin; iq <= qlast; iq++)
     { Jq = (iq + 1) / 2; ouf << ",v" << Jq << ","; }
    ouf << endl; ouf.precision(4);
    for (qb = -basis.maxAbsQntNbr; qb <= basis.maxAbsQntNbr; qb++) {
      ib = qb + basis.maxAbsQntNbr;
      ouf << "c(" << qb << ")";
      for (iq = qbegin; iq <= qlast; iq++) {
        ouf << "," << basis.z[iq][ib].real()
            << "," << basis.z[iq][ib].imag();
      }
      ouf << endl;
    }
    delete ofmt;
    ouf.close();
  }
  return NORM;
}

/* ----- write eigen functions ---------------------------------------- */

int write_eigen_funcs(string &basfln, basisFreeRotor &basis, int limAQn) {
  int iq, Jq, nqn = limAQn * 2 + 1, xdeg;
  int nPages = (nqn - 1) / 125 + 1, jp, qbegin, qlast;
  double xrad;
  complex<double> *egf;
  vector<ofstream *> ouf(nPages);
  ofstream *ouf1;
  string outfln;
  vector<outFormat *> ofmt(nPages);
  outFormat *ofmt1;

  ouf.clear(); ofmt.clear();
  for (jp = 0; jp < nPages; jp++) {
    qbegin = jp * 125; qlast = jp * 125 + 124;
    if (qlast >= nqn) { qlast = nqn - 1; }
    outfln = basfln + "_eigen_funcs" + format(jp, 2, '0') + ".csv";
    ouf1 = new ofstream(outfln.c_str());
    if (!(*ouf1)) { return errms(FAILopen, outfln); }
    ouf.push_back(ouf1);
    ofmt1 = new outFormat(ouf[jp]);
    ofmt.push_back(ofmt1);
    *ouf[jp] << "eigen_functions";
    for (iq = qbegin; iq <= qlast; iq++) { *ouf[jp] << ",(real),(imag)"; }
    *ouf[jp] << "\nx";
    for (iq = qbegin; iq <= qlast; iq++) { Jq = (iq + 1) / 2; *ouf[jp] << ",f" << Jq << ","; }
    *ouf[jp] << endl;
  }
  for (xdeg = 0; xdeg <= 360; xdeg++) {
    xrad = ((double)xdeg / 180.) * M_PI;
    egf = basis.eigenFunc(xrad, limAQn);
    for (jp = 0; jp < nPages; jp++) {
      qbegin = jp * 125; qlast = jp * 125 + 124;
      if (qlast >= nqn) { qlast = nqn - 1; }
      *ouf[jp] << xdeg; ouf[jp]->precision(4);
      for (iq = qbegin; iq <= qlast; iq++)
        { *ouf[jp] << "," << egf[iq].real() << "," << egf[iq].imag(); }
      *ouf[jp] << endl;
    }
  }
  for (jp = 0; jp < nPages; jp++) {
    delete ofmt[jp]; ouf[jp]->close(); delete ouf[jp];
  }
  return NORM;
}

/* ----- calculate and write partition functions ---------------------- */

int write_part_funcs(string &basfln, basisFreeRotor &basis, int limAQn,
 tempList &tpls) {
  int itl, ntl = tpls.size();
  int iq, Jq, symn = basis.nSpinW.size(), iss, sumSw = 0;
  double Tmax, TmaxInL, T, errPf, kT, Qpf;
  ofstream ouf;
  string outfln = basfln + "_part_funcs.csv";
  outFormat *ofmf;

  for (iss = 0; iss < symn; iss++) { sumSw += basis.nSpinW[iss]; }
  Tmax = basis.w[limAQn * 2] * KELWAVN / log(100.);
  TmaxInL = tpls.maxLowerEq(Tmax);
  if (TmaxInL < 0.) { return errms(LRGQerr); }
  errPf = exp(-basis.w[limAQn * 2] * KELWAVN / TmaxInL);
  cout << "estimated partition function error is " << errPf << " at "
   << TmaxInL << " K.\n";
  if (TmaxInL < tpls.maxTemp()) { errms(LMTQrng); }

  ouf.open(outfln.c_str());
  if (!ouf) { return errms(FAILopen, outfln); }
  ofmf = new outFormat(&ouf);
  ouf.precision(8);
  ouf << "10000/T,T,Q\n";
  for (itl = 0; itl < ntl; itl++) {
    T = tpls[itl];
    if (T <= Tmax) {
      kT = T / KELWAVN; Qpf = 0.;
      for (iq = (limAQn * 2); iq >= 0; iq--) {
        Jq = (iq + 1) / 2; iss = Jq % symn;
        Qpf += (double)(basis.nSpinW[iss]) * exp(-basis.w[iq] / kT);
      }
      Qpf /= (double)sumSw;
      ouf << (10000. / T) << "," << T << "," << Qpf << endl;
    }
  }
  delete ofmf;
  ouf.close();
  return NORM;
}

/* ----- error message functions -------------------------------------- */

int errms(errcode ec, string ms) {
  string s;
  switch(ec) {
    case ERRusage: s = "usage: bx1HRsol basefilename"; break;
    case FAILopen: s = "- Failed to open [" + ms + "]."; break;
    case READerr: s = "- File read error."; break;
    case UNKNkey: s = ".. Ignored unknown key [" + ms + "]."; break;
    case NOmaqn: s = "- No maxAbsQN input found."; break;
    case INVmaqn: s = "- Invalid maxAbsQN input."; break;
    case NOrotc: s = "- No rotConst input found."; break;
    case INVrotc: s = "- Invalid rotConst input."; break;
    case INVsymn: s = "- Invalid symNumbr input."; break;
    case INVnuss: s = "- Invalid nuSpinSt input."; break;
    case INVpotp: s = "- Invalid potPars input."; break;
    case NOVpotp: s = "- No valid potPars input found."; break;
    case NUMerr: s = "- Numerical error."; break;
    case BADconv: s = "warning: Bad convergence."; break;
    case FATALconv: s = "- Extremely bad convergence. Fatal."; break;
    case LRGQerr: s = "- No partfunc calc can be performed."; break;
    case LMTQrng: s = "warning: Limited T-range for Q-calc."; break;
  }
  cout << s << endl;
  return ERRE;
}

