/* ======================================================================
     "bx1fitPlls.cc"
       - Fit a potential energy function by Linear Least Squares method
          For generation of an input for bx1VIBsol - basis set expansion
          solver for 1D Schroedinger equation (PS/H-G)
             Copyright (c) 2005       by A. Miyoshi, Univ. Tokyo
                         created: Sept. 15, 2005
                          edited: Sept. 18, 2005   Rev. 0.10
                     last edited: Oct.  20, 2005   Rev. 0.11
 ------------------------------------------------------------------------
 - Rev. 0.11 : add function to use derivatives as data point
               using multi-vars LLS
 ------------------------------------------------------------------------
 usage:
  $ bx1fitPlls basefilename
  - input is read from 'basefilename'.fit

 ====================================================================== */

#include "bx1fitPlls.h"

/* ===== main function ================================================ */

int main(int argc, char *argv[]) {
  int optPS = false, optFS = false;
  string basfln, option;
  vector<string> fnforms;
  multiVarLLS mLLS;
  vector<double> mlParam, sdParam, ffncCoef;
  vector<vector<double> > ccParam;
  powFourSer potp;
  vector<int> degDer;
  contInput cnt;

  if (argc != 2) { errms(ERRusage); return 1; }
  basfln = argv[1];
  if (read_input(basfln, mLLS, degDer, cnt, optPS, optFS) == ERRE)
   { return 1; }
  if (mLLS.solve(mlParam, sdParam, ccParam) == ERRE) { return 1; }
  fnforms = mLLS.funcsStrAll(); ffncCoef = mLLS.getFixCoef();
  reform_funcstr(fnforms, degDer);
  write_results(mlParam, sdParam, ccParam, fnforms, ffncCoef);
  if (optPS) {
    if(mLLS.isPowerSer(0, potp)) {
      if (write_potpars(basfln, potp, cnt) == ERRE) { return 1; }
    } else { errms(NOTps); }
  } else if (optFS) {
    if(mLLS.isFourierSer(0, potp)) {
      if (write_potpars(basfln, potp, cnt) == ERRE) { return 1; }
    } else { errms(NOTfs); }
  }
  return 0;
}

/* ----- read input --------------------------------------------------- */

int read_input(string &basfln, multiVarLLS &mLLS, vector<int> &dgd,
 contInput &cnt, int &ops, int &ofs) {
  int ndat = 0, blkend, numFuncs, deg, ifnc;
  ifstream inf;
  string inpfln = basfln + ".fit", key, line, s;
  vector<string> vs, vsv;
  listFnUnits opFnc, fxFnc;
  vector<double> fxCf;
  map<int, dataPntsXYW> dSet;
  map<int, dataPntsXYW>::iterator pds;
  dataPntsXYW nullDpnts;
  pointXYW dpnt;

  cnt.clear();
  inf.open(inpfln.c_str());
  if (!inf) { return errms(FAILopen, inpfln); }
  opFnc.clear(); fxFnc.clear(); fxCf.clear(); dSet.clear();
  nullDpnts.clear(); dgd.clear();
  while (inpTokenKeyValues(inf, key, vs) == NORM) {
    if (key == "#") {
      cnt.title.push_back(vs[0]);
    } else if (key == "outputOption") {
      ops = false; ofs = false;
      if (vs[0] == "powerSeries") { ops = true; }
      else if (vs[0] == "FourierSeries") { ofs = true; }
    } else if (key == "continue{") {
      blkend = false;
      while (!blkend) {
        if (getLineNonBlank(inf, line) == ERRE) { return errms(READerr); }
        if (line[0] == '}') { blkend = true; }
        else { cnt.contPar.push_back(line); }
      }
    } else if (key == "optFuncs{") {
      blkend = false;
      while (!blkend) {
        if (inpToken(inf, vsv) == ERRE) { return errms(READerr); }
        if (vsv[0][0] == '}') { blkend = true; }
        else {
          if (opFnc.addFrom(vsv) == ERRE) { return errms(INVfunc); }
        }
      }
    } else if (key == "fixFuncs{") {
      blkend = false;
      while (!blkend) {
        if (inpToken(inf, vsv) == ERRE) { return errms(READerr); }
        if (vsv[0][0] == '}') { blkend = true; }
        else {
          fxCf.push_back(atof(vsv[0].c_str()));
          vsv.erase(vsv.begin());
          if (fxFnc.addFrom(vsv) == ERRE) { return errms(INVfunc); }
        }
      }
    } else if (key == "xyTable{") {
      blkend = false;
      while (!blkend) {
        if (inpToken(inf, vsv) == ERRE) { return errms(READerr); }
        if (vsv[0][0] == '}') { blkend = true; }
        else if (vsv.size() < 2) { return errms(INVdata); }
        else {
          deg = 0;
          if (vsv[0].substr(0, 5) == "deriv") {
            s = vsv[0].substr(5);
            if (s.length() == 0) { deg = 1; }
            else {
              deg = atoi(s.c_str());
              if (deg < 0) { return errms(INVdegd); }
              if (deg == 0) { deg = 1; }
            }
            vsv.erase(vsv.begin());
            if (vsv.size() < 2) { return errms(INVdata); }
          }
          dpnt.x = atof(vsv[0].c_str());
          dpnt.y = atof(vsv[1].c_str());
          dpnt.w = 1.;
          if (vsv.size() > 2) { dpnt.w = atof(vsv[2].c_str()); }
          dSet[deg].push_back(dpnt);
          ndat++;
        }
      }
    } else { errms(UNKNkey, key); }
  }
  inf.close();
  if (dSet.find(0) == dSet.end())
   { dSet.insert(make_pair(0, nullDpnts)); }
  for (pds = dSet.begin(); pds != dSet.end(); pds++)
   { dgd.push_back(pds->first); }
  numFuncs = dgd.size();
  mLLS.init(numFuncs);
  mLLS.set1OptFnc(0, opFnc); mLLS.set1FixFncCoef(0, fxFnc, fxCf);
  for (ifnc = 1; ifnc < numFuncs; ifnc++)
   { mLLS.setDerivFunc(ifnc, 0, dgd[ifnc]); }
  if (!mLLS.isFuncSet()) { return errms(NOOPTfunc); }
  if (ndat < 1) { return errms(NOXYdata); }
  for (ifnc = 0; ifnc < numFuncs; ifnc++) {
    mLLS.setPoints(ifnc, dSet[dgd[ifnc]]);
  }
  return NORM;
}

/* ----- reformat function strings ------------------------------------ */

void reform_funcstr(vector<string> &fnforms, vector<int> &dgd) {
  int nsf = fnforms.size(), isf, deg;
  string fhead;
  for (isf = 0; isf < nsf; isf++) {
    fhead = "y";
    deg = dgd[isf];
    while (deg > 0) {
      if (deg > 1) { fhead += '"'; deg -= 2; }
      else { fhead += "'"; deg--; }
    }
    fnforms[isf] = fhead + fnforms[isf];
  }
}

/* ----- write LLS results -------------------------------------------- */

void write_results(vector<double> &mlp, vector<double> &sdp,
 vector<vector<double> > &pcc, vector<string> &fnforms,
 vector<double> &ffc) {
  int npars = mlp.size(), nffnc = ffc.size(), is, ip, iq, jff;
  string s;
  outFormat ofmt(&cout);

  cout << " ===== results of linear least squares regression =====\n";
  cout << " Function:\n";
  for (is = 0; is < fnforms.size(); is++) {
    cout << "  " << fnforms[is] << endl;
  }
  cout.precision(7); cout.setf(ios::scientific);
  if (nffnc > 0) {
    cout << " Constants:\n";
    for (jff = 0; jff < nffnc; jff++) {
      cout << "  c" + format(jff) + " = ";
      cout.width(15); cout << ffc[jff] << endl;
    }
  }
  cout << " Optimum parameters (+-) standard deviations (relative)\n";
  for (ip = 0; ip < npars; ip++) {
    cout << "  a" + format(ip) + " = ";
    cout.width(15); cout << mlp[ip] << " (+-) ";
    cout.width(15); cout << sdp[ip] << " (";
    cout.unsetf(ios::scientific); cout.setf(ios::fixed);
    cout.precision(2); cout.width(7);
    if (fabs(mlp[ip]) < VAL0THR) {
      cout << "   -   ";
    } else {
      cout << fabs(sdp[ip] / mlp[ip]) * 100;
    }
    cout << " %)\n";
    cout.unsetf(ios::fixed); cout.setf(ios::scientific);
    cout.precision(7);
  }
  cout << " Parameter correlation coefficients\n";
  cout << "     ";
  for (iq = 0; iq < npars - 1; iq++)
   { s = "       a" + format(iq); cout << s.substr(s.length() - 9); }
  cout << endl;
  cout.unsetf(ios::scientific); cout.setf(ios::fixed); cout.precision(5);
  for (ip = 1; ip < npars; ip++) {
    s = "  a" + format(ip) + " "; cout << s.substr(0, 5);
    for (iq = 0; iq < ip; iq++) { cout.width(9); cout << pcc[ip][iq]; }
    cout << endl;
  }
}

/* ----- write potential parameters (for bx1VIBsol &) ----------------- */

int write_potpars(string &basfln, powFourSer &pp, contInput &cnt) {
  ofstream ouf;
  string outfln = basfln + ".inp";
  outFormat *ofmt;
  int ict, nct = cnt.title.size(), icp, ncp = cnt.contPar.size();
  int npotp = pp.size(), ip;

  ouf.open(outfln.c_str());
  if (!ouf) { return errms(FAILopen, outfln); }
  ofmt = new outFormat(&ouf);
  for (ict = 0; ict < nct; ict++)
   { ouf << "# " << cnt.title[ict] << endl; }
  ouf << endl;
  for (icp = 0; icp < ncp; icp++) { ouf << cnt.contPar[icp] << endl; }
  ouf << "\npotPars{           ! potential parameters\n";
  ouf.precision(8);
  for (ip = 0; ip < npotp; ip++) {
    ouf << "  " << pp[ip].order << "   " << (pp[ip].coef * pp[ip].param)
     << endl;
  }
  ouf << "}\n";
  delete ofmt;
  ouf.close();
  return NORM;
}

/* ----- error message functions -------------------------------------- */

int errms(errcode ec, string ms) {
  string s;
  switch(ec) {
    case ERRusage: s = "usage: bx1fitPlls basefilename"; break;
    case INVopt: s = "invalid option [" + ms + "] ignored."; break;
    case FAILopen: s = "- Failed to open [" + ms + "]."; break;
    case READerr: s = "- File read error."; break;
    case INVdegd: s = "- Invalid degree of derivative."; break;
    case INVfunc: s = "- Invalid function input."; break;
    case INVdata: s = "- Invalid data input."; break;
    case NOOPTfunc: s = "- No valid optFuncs found."; break;
    case NOXYdata: s = "- No valid xyTable found."; break;
    case UNKNkey: s = ".. Ignored unknown key [" + ms + "]."; break;
    case NOTps: s = "- function is NOT powerSeries."; break;
    case NOTfs: s = "- function is NOT FourierSeries."; break;
  }
  cout << s << endl;
  return ERRE;
}

